import re

import pandas as pd
import sympy as sp

from ._safeguards import _cached_caller_globals, create_key, retrieve_passkey
from .dgcv_core import clearVar, listVar, variableProcedure
from .dgcv_formatter import get_style
from .finite_dim_algebras import AlgebraElement, FAClass
from .tensors import tensorProduct


class Tanaka_symbol(sp.Basic):
    """
    DGCV class representing a symbol-like object for Tanaka prolongation.

    Parameters
    ----------
    GLA : FAClass, (with negatively graded Lie algebra structure)
        ambiant Lie algebra's negative part. The first entry in GLS.gading must be a list of negative weights, which will be used for the prolongation degrees

    Methods
    -------
    prolong

    Examples
    --------
    """
    def __new__(cls, GLA, nonnegParts,assume_FGLA = True, subspace = None, index_threshold = None):

        if not isinstance(GLA, FAClass):
            raise TypeError(
                "`Tanaka_symbol` expects `GLA` to be a graded Lie algebra (`FAClass` in particular) with negative weights in the first element of `GLA.grading`."
            )
        elif isinstance(GLA.grading[0],(list,tuple)) and not all(j<0 for j in GLA.grading[0]):
            raise TypeError(
                "`Tanaka_symbol` expects `GLA` to be a graded Lie algebra (`FAClass` in particular) with negative weights in the first element of `GLA.grading`."
            )
        elif isinstance(GLA.grading,(int,float,sp.Expr)) and not all(j<0 for j in GLA.grading):
            raise TypeError(
                "`Tanaka_symbol` expects `GLA` to be a graded Lie algebra (`FAClass` in particular) with negative weights in the first element of `GLA.grading`."
            )

        if subspace:
            if not isinstance(subspace,(list,tuple)):
                raise TypeError(
                    "`Tanaka_symbol` expects `subpsace` to be a list of AlgebraElement instances belonging to the `FAClass` `GLA`."
                )
            if not all(isinstance(j,AlgebraElement) for j in subspace) or not all(j.algebra==GLA for j in subspace):
                raise TypeError(
                    "`Tanaka_symbol` expects `subpsace` to be a list of AlgebraElement instances belonging to the `FAClass` given for `GLA`."
                )
        else:
            subspace = GLA

        if not isinstance(nonnegParts,(list,tuple)):
            raise TypeError(
                "`Tanaka_symbol` expects `nonnegParts` to be a list of `tensorProduct` built from the `FAClass` given for `GLA` with `valence` of the form (1,0,...0)."
            )
        if not all(isinstance(j,tensorProduct) for j in nonnegParts) or not all(j.vector_space==GLA for j in nonnegParts):
            raise TypeError(
                "`Tanaka_symbol` expects `nonnegParts` to be a list of `tensorProduct` instances built from the `FAClass` given for `GLA` with `valence` of the form (1,0,...0)."
            )

        def valence_check(tp):
            for j in tp.coeff_dict:
                valence = j[len(j)//2:]
                if valence[0] != 1:
                    return False
                if not all(j in {0} for j in valence[1:]):
                    return False
            return True

        if not all(valence_check(j) for j in nonnegParts):
            raise TypeError(
                "`Tanaka_symbol` expects `nonnegParts` to be a list of `tensorProduct` instances built from the `FAClass` given for `GLA` with `valence` of the form (1,0,...0)."
            )

        obj = sp.Basic.__new__(cls, GLA, nonnegParts)

        obj._subspace = subspace

        return obj

    def __init__(self, GLA, nonnegParts, assume_FGLA = True, subspace = None, index_threshold = None):
        self.ambiantGLA = GLA
        self.negativePart = self._subspace
        if not assume_FGLA:
            raise KeyError('A non FGLA `Tanaka_symbol` class has not be developed yet...')
        self.assume_FGLA = assume_FGLA
        self.nonnegParts = nonnegParts
        negWeights = sorted(tuple(set(GLA.grading[0]))) if isinstance(GLA.grading,(list,tuple)) else sorted(tuple(set(GLA.grading)))
        if negWeights[-1]!=-1:
            raise AttributeError('`Tanaka_symbol` expects negatively graded LA to have a weight -1 component.')
        # if not all(isinstance(j,int) for j in negWeights):
        #     raise AttributeError('`Tanaka_symbol` expects negatively graded LA to have a weight -1 component.')#!!!
        self.negWeights = negWeights
        nonNegWeights = sorted(tuple(set([j.compute_weight()[0] for j in nonnegParts])))
        self.height = nonNegWeights[-1]
        self.depth = negWeights[0]
            #!!!
        # if nonNegWeights[0]!=0 or not isinstance(nonNegWeights[-1],int) or len(nonNegWeights)!=self.height +1:
        #     raise TypeError('`Tanaka_symbol` expects the given `nonnegParts` list to have only tensorPower instances with non-negative primary weight, whose primary weight comprise a list of consecutive integers containing 0.')
        self.weights = negWeights+nonNegWeights
        GLA_levels = dict()
        for weight in negWeights:
            level = [j for j in GLA.basis if j.check_element_weight()[0]==weight]
            GLA_levels[weight]=level
        self.GLA_levels = GLA_levels

        nonneg_levels = dict()
        for weight in nonNegWeights:
            level = [j for j in nonnegParts if j.compute_weight()[0]==weight]
            nonneg_levels[weight]=level
        self.nonneg_levels = nonneg_levels
        levels = self.GLA_levels | self.nonneg_levels

        class dynamic_dict(dict):                   # special dict structure for the graded decomp.
            def __init__(self, dict_data, initial_index = None):
                super().__init__(dict_data)
                self.index_threshold = initial_index

            def __getitem__(self, key):
                # If index_threshold is None, behave like a regular dictionary
                if self.index_threshold is None:
                    return super().get(key, None)

                # Otherwise, apply the threshold logic
                if isinstance(key, int) and key >= self.index_threshold:
                    return []  # Return an empty list for keys > index_threshold
                return super().get(key, None)  # Default behavior otherwise

            def _set_index_thr(self, new_threshold):
                # Allow None or an integer as valid values
                if not (isinstance(new_threshold, (int,float,sp.Expr)) or new_threshold is None):
                    raise TypeError("index_threshold must be an integer or None.")
                self.index_threshold = new_threshold
        self._GLA_structure = dynamic_dict
        self.levels = dynamic_dict(levels, initial_index = index_threshold)

        self._test_commutators = None

    @property
    def test_commutators(self):
        if self._test_commutators:
            return self._test_commutators
        if self.assume_FGLA:
            deeper_levels = sum([self.GLA_levels[j] for j in self.negWeights[:-1]],[])
            f_level = self.GLA_levels[-1]
            first_commutators = [(f_level[j],f_level[k],f_level[j]*f_level[k]) for j in range(len(f_level)) for k in range(j+1,len(f_level))]
            remaining_comm = [(j,k,j*k) for j in f_level for k in deeper_levels]
            self._test_commutators = first_commutators+remaining_comm
            return first_commutators+remaining_comm
        else:
            pass

    def _prolong_by_1(self, levels, height): # start with self.levels, self.height (height must match levels structure)
        if self.assume_FGLA:
            if len(levels[height])==0:
                new_levels = levels
                print(f'height: {height}, type: {type(height)}')
                new_levels._set_index_thr(height)
                stable = True
            else:
                ambiant_basis = []
                for weight in self.negWeights:
                    ambiant_basis += [k@(j.dual()) for j in self.GLA_levels[weight] for k in levels[height+1+weight]]

                varLabel=create_key(prefix="center_var")   # label for temparary variables
                variableProcedure(
                    varLabel,
                    len(ambiant_basis),
                    _tempVar=retrieve_passkey()
                )
                tVars = _cached_caller_globals[varLabel]    # pointer to tuple of coef vars

                general_elem = sum([tVars[j]*ambiant_basis[j] for j in range(1, len(tVars))],tVars[0]*ambiant_basis[0])
                eqns = []
                for triple in self.test_commutators:
                    derivation_rule = (general_elem*triple[0])*triple[1]+triple[0]*(general_elem*triple[1])-general_elem*triple[2]
                    if isinstance(derivation_rule,tensorProduct):
                        eqns += list(derivation_rule.coeff_dict.values())
                    elif isinstance(derivation_rule,AlgebraElement):
                        eqns += derivation_rule.coeffs
                solution = sp.solve(eqns,tVars,dict=True)
                if not solution:
                    raise RuntimeError('`Tanaka_symbol.prolongation` failed at a step where sympy.solve was being applied.')
                el_sol = general_elem.subs(solution[0])

                free_variables = tuple(set.union(*[set(sp.sympify(j).free_symbols) for j in el_sol.coeff_dict.values()]))

                return_list = []
                for var in free_variables:
                    basis_element = el_sol.subs({var: 1}).subs(
                        [(other_var, 0) for other_var in free_variables if other_var != var]
                    )
                    return_list.append(basis_element)

                clearVar(*listVar(temporary_only=True), report=False)

                new_levels =  self._GLA_structure(levels | {height+1:return_list}, levels.index_threshold)
                stable = False
            return new_levels, stable

        else:
            levels, True ### Edit, TO DO!!!

    def prolong(self, iterations, return_symbol = False):
        if not isinstance(iterations,int) or iterations<1:
            raise TypeError('`prolong` expects `iterations` to be a positive int.')
        levels = self.levels
        height = self.height
        stable = False
        for j in range(iterations):
            if stable:
                break
            levels, stable = self._prolong_by_1(levels,height)
            height += 1
        if return_symbol:
            new_nonneg_parts = []
            for key,value in levels.items():
                if key>=0:
                    new_nonneg_parts += value
            return Tanaka_symbol(self.ambiantGLA,new_nonneg_parts,assume_FGLA=self.assume_FGLA,index_threshold=levels.index_threshold)
        else:
            return levels

    def summary(self, style=None, use_latex=False):
        """
        Generates a pandas DataFrame summarizing the Tanaka_symbol data, with optional styling and LaTeX rendering.

        Parameters
        ----------
        style : str, optional
            A string key to retrieve a custom pandas style from the style_guide.
        use_latex : bool, optional
            If True, formats the table with rendered LaTeX in the Jupyter notebook.
            Defaults to False.

        Returns
        -------
        pandas.DataFrame or pandas.io.formats.style.Styler
            A styled DataFrame summarizing the Tanaka_symbol data, optionally with LaTeX rendering.
        """

        # Prepare data for the DataFrame
        data = {
            "Weight": [],
            "Dimension": [],
            "Basis": [],
        }
        for weight, basis in self.levels.items():
            data["Weight"].append(weight)
            data["Dimension"].append(len(basis))
            if use_latex:
                # Render LaTeX for the basis elements
                latex_basis = ", ".join(f"${sp.latex(b)}$" for b in basis)
                data["Basis"].append(latex_basis)
            else:
                data["Basis"].append(", ".join(map(str, basis)))

        # Create the DataFrame
        df = pd.DataFrame(data)
        df = df.sort_values(by="Weight").reset_index(drop=True)

        # Retrieve the custom style
        if style is not None:
            pandas_style = get_style(style)
        else:
            pandas_style = get_style("default")

        # Dynamically extract relevant styles
        border_color = "#ccc"  # Default border color
        hover_background = None
        hover_color = None

        # Extract border color from the "table" styles
        if "table" in pandas_style and "props" in pandas_style["table"]:
            for prop in pandas_style["table"]["props"]:
                if "border" in prop[0]:  # Check if the property is 'border'
                    # Extract the color using regex for flexible matching
                    match = re.search(r"#(?:[0-9a-fA-F]{3}){1,2}", prop[1])
                    if match:
                        border_color = match.group(0)  # Extract the hex color

        # Extract border color if defined
        if "table" in pandas_style and "props" in pandas_style["table"]:
            for prop in pandas_style["table"]["props"]:
                if "border" in prop[0]:
                    border_color = prop[1].split()[-1]  # Extract last part as color

        # Extract hover styles if defined
        if "hover" in pandas_style and "props" in pandas_style["hover"]:
            for prop in pandas_style["hover"]["props"]:
                if "background-color" in prop[0]:
                    hover_background = prop[1]
                if "color" in prop[0]:
                    hover_color = prop[1]

        # Define additional styles for column separators and header line
        additional_styles = [
            {"selector": "th", "props": [("border-bottom", f"2px solid {border_color}"), ("text-align", "center")]},
            {"selector": "td", "props": [("border-right", f"1px solid {border_color}"), ("text-align", "center")]},
        ]

        # Apply hover styles to column separators, if defined
        if hover_background or hover_color:
            additional_styles.append({
                "selector": "td:hover",
                "props": [
                    ("border-right", f"1px solid {hover_background or border_color}"),
                    ("background-color", hover_background),
                    ("color", hover_color),
                ],
            })

        # Merge custom styles with additional styles
        table_styles = pandas_style + additional_styles

        # Apply styles
        if use_latex:
            # Convert the DataFrame to HTML with LaTeX rendering
            styled_df = (
                df.style
                .hide(axis="index")  # Suppress the index column first
                .format({"Basis": lambda x: f"<div style='text-align: center;'>{x}</div>"})
                .set_caption("Summary of Tanaka Symbol (with prolongations)")
                .set_table_styles(table_styles)
            )
            return styled_df

        # Apply styles for the non-LaTeX version
        styled_df = (
            df.style
            .hide(axis="index")  # Suppress the index column first
            .set_caption("Summary of Tanaka Symbol (with prolongations)")
            .set_table_styles(table_styles)
        )

        return styled_df

    def __str__(self):
        result = ["Tanaka Symbol:"]
        result.append("Graded vector space with weights and dimensions:")
        for weight, basis in self.levels.items():
            dim = len(basis)
            basis_str = ", ".join(map(str, basis))
            result.append(f"  Weight {weight}: Dimension {dim}")
            result.append(f"    Basis: [{basis_str}]")
        return "\n".join(result)

    def _repr_latex_(self):
        lines = [
            r"\text{Tanaka Symbol:}\\",
            r"\text{Graded vector space with weights and dimensions:}\\",
            r"\begin{alignedat}{2}"
        ]
        for weight, basis in self.levels.items():
            dim = len(basis)
            basis_latex = ", ".join(map(lambda b: sp.latex(b), basis))
            lines.append(
                rf"&\text{{Weight {weight}: }} &\text{{Dimension {dim}, Basis: }} [{basis_latex}] \\"
            )
        lines.append(r"\end{alignedat}")
        return "$\n" + "\n".join(lines) + "\n$"

    def _sympystr(self, printer):
        result = ["Tanaka Symbol:"]
        result.append("Weights and Dimensions:")
        for weight, basis in self.levels.items():
            dim = len(basis)
            basis_str = ", ".join(printer.doprint(b) for b in basis)
            result.append(f"  {weight}: Dimension {dim}, Basis: [{basis_str}]")
        return "\n".join(result)

    def __repr__(self):
        return f"Tanaka_symbol(ambiantGLA={repr(self.ambiantGLA)}, levels={len(self.levels)} levels)"



"""
DGCV: Differential Geometry with Complex Variables

This module provides various combinatorial functions used throughout the DGCV package, 
primarily focusing on efficient computation of Cartesian products, permutations, and 
related operations. These functions are optimized for performance and designed to 
handle large inputs by utilizing generators where possible.

Key Functions
-------------
- carProd: Computes the Cartesian product of multiple lists.
- carProdWithOrder: Computes the Cartesian product while removing equivalent permutations.
- carProdWithoutRepl: Computes the Cartesian product, excluding repeated elements.
- chooseOp: Generates combinations with optional order, replacement, and homogeneity degree filtering.
- permSign: Calculates the signature of a permutation.
- permuteTupleEntries: Applies a permutation to the entries of a tuple or list.
- permuteTuple: Applies a permutation to reorder a tuple or list.
- permuteArray: Applies a permutation to the indices of a k-dimensional array.
- alternatingPartOfArray: Computes the alternating part of a multilinear operator represented by an array.

Dependencies
------------
- sympy: For arrays (MutableDenseNDimArray) and mathematical operations.

Notes
-----
This module minimizes dependencies and is critical for handling combinatorics within the DGCV package's computationally intensive operations.

"""

############## dependencies
from sympy import MutableDenseNDimArray


############## general combinatorics
def carProd(*args):
    """
    Compute the Cartesian product of a variable number of lists.

    This function takes multiple lists as input and computes their
    Cartesian product, yielding tuples containing elements from each list.

    Parameters
    ----------
    *args : lists
        The input lists whose Cartesian product is to be computed.

    Returns
    -------
    generator
        A generator yielding tuples representing the Cartesian product.

    Examples
    --------
    >>> list(carProd([1, 2], [3, 4]))
    [(1, 3), (1, 4), (2, 3), (2, 4)]

    Notes
    -----
    This function avoids loading the entire Cartesian product into memory
    at once by using generators.

    Raises
    ------
    TypeError
        If any of the input arguments are not iterable.
    """

    def carProdTwo(arg1, arg2):
        return (j + (k,) for j in arg1 for k in arg2)

    if len(args) == 1:
        return ((j,) for j in args[0])
    else:
        resultLoc = ((j,) for j in args[0])
        for j in range(1, len(args)):
            resultLoc = carProdTwo(resultLoc, args[j])
        return resultLoc

def carProd_with_weights_without_R(*args):
    """
    Form cartesian product (filtered for replacement) of a variable number of lists whose elements are marked with a weight (specifically, list entries should be length 2 lists whose first element goes into the car. prod. space and second element a scalar vaule). Weights are multiplied when elements are joined into a list (i.e., element of the cartesian product space).

    Args:
        args: List

    Returns: List
        list of lists marked with weights. Specifically, a list of length 2 lists, each conaintaining a scalar (e.g., number or sympy.Expr) in the second position representing a weight and a list representing the car. prod. element

    Raises:
    """

    def prodOfTwo(arg1, arg2):
        return [
            [j[0] + (k[0],), j[1] * k[1]]
            for j in arg1
            for k in arg2
            if len(set(j[0] + (k[0],))) == len(j[0] + (k[0],))
        ]

    if len(args) == 1:
        return [[(j[0],), j[1]] for j in args[0]]
    else:
        resultLoc = ([(j[0],), j[1]] for j in args[0])
        for j in range(len(args) - 1):
            resultLoc = prodOfTwo(resultLoc, list(args[j + 1]))
        return resultLoc

def carProdWithOrder(*args):
    """
    Compute the Cartesian product of lists, excluding permutations.

    This function computes the Cartesian product of multiple lists and
    removes elements that are equivalent up to permutation. The input
    lists are pre-sorted to optimize efficiency, and the function yields
    unique combinations lazily.

    Parameters
    ----------
    *args : lists
        The input lists whose Cartesian product is to be computed.

    Returns
    -------
    generator
        A generator yielding unique tuples representing the Cartesian
        product, with permutations removed.

    Examples
    --------
    >>> list(carProdWithOrder([1, 2], [2, 3]))
    [(1, 2), (1, 3), (2, 3)]

    Notes
    -----
    The input lists are sorted to ensure permutations are excluded more
    efficiently. The function yields results lazily for improved memory
    efficiency.

    Raises
    ------
    TypeError
        If any of the input arguments are not iterable.
    """
    sorted_args = [sorted(arg) for arg in args]
    seen = set()

    for combo in carProd(*sorted_args):
        sorted_combo = tuple(combo)
        if sorted_combo not in seen:
            seen.add(sorted_combo)
            yield sorted_combo

def carProdWithoutRepl(*args):
    """
    Compute Cartesian product excluding repeated elements.

    This function computes the Cartesian product of multiple lists and
    filters out tuples that contain repeated values. The function yields
    tuples lazily for improved memory efficiency.

    Parameters
    ----------
    *args : lists
        The input lists whose Cartesian product is to be computed.

    Returns
    -------
    generator
        A generator yielding tuples that do not contain repeated values.

    Examples
    --------
    >>> list(carProdWithoutRepl([1, 2], [2, 3]))
    [(1, 2), (2, 3)]

    Notes
    -----
    This function is memory efficient and excludes tuples with repeated
    elements in an on-the-fly manner.

    Raises
    ------
    TypeError
        If any of the input arguments are not iterable.
    """
    return (j for j in carProd(*args) if len(set(j)) == len(j))

def carProdWithOrderWithoutRepl(*args):
    """
    Compute Cartesian product excluding permutations and repeated elements.

    This function computes the Cartesian product of multiple lists, filters
    out tuples that contain repeated values, and removes elements equivalent
    up to permutation by sorting the input lists upfront. The function yields
    unique tuples lazily for improved memory efficiency.

    Parameters
    ----------
    *args : lists
        The input lists whose Cartesian product is to be computed.

    Returns
    -------
    generator
        A generator yielding unique tuples that do not contain repeated
        elements or permutations.

    Examples
    --------
    >>> list(carProdWithOrderWithoutRepl([1, 2], [2, 3]))
    [(1, 2), (1, 3)]

    Notes
    -----
    By sorting the input lists beforehand and applying filters during the
    Cartesian product computation, this function minimizes the need to
    process permutations and repetitions separately.

    Raises
    ------
    TypeError
        If any of the input arguments are not iterable.
    """
    sorted_args = [sorted(arg) for arg in args]
    seen = set()

    for combo in carProd(*sorted_args):
        if len(set(combo)) == len(combo):  # Filter out tuples with repeated values
            sorted_combo = tuple(
                combo
            )  # Input is already sorted, no need to re-sort here
            if sorted_combo not in seen:
                seen.add(sorted_combo)
                yield sorted_combo

def chooseOp(
    arg1, arg2, withOrder=False, withoutReplacement=False, restrictHomogeneity=None
):
    """
    Generate all possible combinations of length *arg2* containing elements from *arg1*.

    The function can apply several filters: excluding permutations, preventing duplicate elements,
    and restricting combinations to those with a specified homogeneity degree (sum of elements).

    Parameters
    ----------
    arg1 : list
        The list of elements from which combinations are drawn.
    arg2 : int
        The length of the combinations to be generated.
    withOrder : bool, optional
        If True, removes equivalent combinations that are permutations of each other.
    withoutReplacement : bool, optional
        If True, prevents duplicate elements within a combination.
    restrictHomogeneity : int, optional
        If set, filters combinations to only include those whose elements sum to the given value.

    Returns
    -------
    generator
        A generator yielding tuples of the specified combinations.

    Examples
    --------
    >>> list(chooseOp([1, 2], 2, withOrder=True))
    [(1, 2)]

    >>> list(chooseOp([1, 2, 3], 2, restrictHomogeneity=4))
    [(1, 3), (3, 1), (2, 2)]

    Notes
    -----
    - The `withOrder` and `withoutReplacement` options control whether permutations
      and duplicate elements are included.
    - If `restrictHomogeneity` is set, only tuples whose elements sum to the
      specified value will be returned.

    Raises
    ------
    TypeError
        If the arguments are not in the correct format.
    """

    arg1 = [list(arg1)]

    # Determine which Cartesian product function to use
    if withOrder:
        if withoutReplacement:
            resultLoc = carProdWithOrderWithoutRepl(*arg2 * arg1)
        else:
            resultLoc = carProdWithOrder(*arg2 * arg1)
    else:
        if withoutReplacement:
            resultLoc = carProdWithoutRepl(*arg2 * arg1)
        else:
            resultLoc = carProd(*arg2 * arg1)

    # Apply homogeneity filter if needed
    if isinstance(restrictHomogeneity, int):
        return (j for j in resultLoc if sum(j) == restrictHomogeneity)
    else:
        return resultLoc

def permSign(arg1, returnSorted=False, **kwargs):
    """
    Compute the signature of a permutation of list of integers, and sort it.

    Computation is based on the *merge-sort* algorithm described here:
    https://en.wikipedia.org/wiki/Merge_sort

    The signature (or sign) of a permutation is 1 if the permutation is even,
    and -1 if the permutation is odd.

    Parameters
    ----------
    arg1 : list
        A list containing a permutation of sortable elements

    Returns
    -------
    int
        The signature of the permutation, either 1 (even permutation) or -1 (odd permutation).

    Examples
    --------
    >>> permSign([2, 1, 3])
    -1

    >>> permSign([2, 0, 1])
    1

    Raises
    ------
    ValueError
        If the input is not a valid permutation of consecutive integers.
    """

    def merge_sort(permutation):
        # Length 1 or empty lists do not need to be sorted
        if len(permutation) <= 1:
            return permutation, 0

        # For longer lists, divide them into two smaller parts.
        # Most merge-sort documentation says to split in half
        partition = len(permutation) // 2
        left = permutation[:partition]
        right = permutation[partition:]

        # Recursively merge-sort and count permutation parities.
        left_sorted, left_parity = merge_sort(left)
        right_sorted, right_parity = merge_sort(right)

        # merge the sorted left and right parts in a sorted way while counting
        # parity of the permutation from "concatenation" to "sorted" merge.
        merged_list, merge_parity = merge_and_count(left_sorted, right_sorted)

        sorting_parity = left_parity + right_parity + merge_parity

        return merged_list, sorting_parity

    def merge_and_count(left, right):
        # we'll build the sorted merge in a list
        merged = []
        # and count the number of swaps performed as we build it 
        # (starting with parity = 0)
        parity = 0

        # Pull elements from the two lists into the merged list
        # by comparing the first element not yet pulled in from
        # either list and taking the smaller one. Do this until
        # all elements from one list are pulled into merged.
        i = j = 0
        while i < len(left) and j < len(right):
            if left[i] <= right[j]:
                merged.append(left[i])
                i += 1
            else:
                merged.append(right[j])
                j += 1
                # pulling in the leading element from the right list
                # requires swapping it with the remaining elements in the
                # left list. Upate parity accordingly
                parity += (len(left) - i)

        # one of the sublists may not have been exhaust, so add what 
        # remains to the end of the merged list.
        merged.extend(left[i:])
        merged.extend(right[j:])

        return merged, parity


    # Count inversions in the permutation and get the sorted list
    sorted_list, inversions = merge_sort(arg1)

    # Compute the sign based on the number of inversions
    sign = 1 if inversions % 2 == 0 else -1

    # Return based on the returnSorted flag
    if returnSorted:
        return sign, sorted_list
    else:
        return sign

def weightedPermSign(permutation, weights, returnSorted=False, use_degree_attribute=False):
    def merge_sort(permutation, weights):
        # Base case: single element or empty list
        if len(permutation) <= 1:
            return permutation, weights, 0

        # Split into left and right parts
        partition = len(permutation) // 2
        left = permutation[:partition]
        right = permutation[partition:]
        left_weights = weights[:partition]
        right_weights = weights[partition:]

        # Recursively sort and count parities
        left_sorted, left_weights_sorted, left_parity = merge_sort(left, left_weights)
        right_sorted, right_weights_sorted, right_parity = merge_sort(right, right_weights)

        # Merge sorted parts while counting weighted parity
        merged_list, merged_weights, merge_parity = merge_and_count(
            left_sorted, right_sorted, left_weights_sorted, right_weights_sorted
        )

        # Combine parities
        sorting_parity = (left_parity + right_parity + merge_parity) % 2

        return merged_list, merged_weights, sorting_parity


    def merge_and_count(left, right, left_weights, right_weights):
        merged = []
        merged_weights = []
        parity = 0

        i = j = 0
        while i < len(left) and j < len(right):
            if left[i] <= right[j]:
                merged.append(left[i])
                merged_weights.append(left_weights[i])
                i += 1
            else:
                merged.append(right[j])
                merged_weights.append(right_weights[j])
                # Weighted parity calculation
                if use_degree_attribute:
                    parity += (sum([mu.degree for mu in left_weights[i:]]) * (right_weights[j].degree)) % 2
                else:
                    parity += (sum(left_weights[i:]) * right_weights[j]) % 2
                j += 1

        # Append remaining elements
        merged.extend(left[i:])
        merged_weights.extend(left_weights[i:])
        merged.extend(right[j:])
        merged_weights.extend(right_weights[j:])

        return merged, merged_weights, parity

    # Sort and compute weighted parity
    sorted_list, sorted_weights, inversions = merge_sort(permutation, weights)

    # Compute the sign based on inversions
    sign = 1 if inversions % 2 == 0 else -1

    if returnSorted:
        return sign, sorted_list, sorted_weights
    else:
        return sign

def shufflings(list1: list | tuple, list2: list | tuple):
    """
    Yield all order-preserving shufflings of list1 and list2.

    This is achieved by recursively building a tree of incrementally longer lists,
    starting from the empty list []. Each step appends the next unused element 
    from either list1 or list2, preserving the relative order within each list.

    Parameters
    ----------
    list1 : list or tuple
        First sequence to merge.
    list2 : list or tuple
        Second sequence to merge.

    Yields
    ------
    list
        A single shuffling of list1 and list2.

    Examples
    --------
    >>> list(shufflings([1, 2], ['a', 'b']))
    [[1, 2, 'a', 'b'], [1, 'a', 2, 'b'], [1, 'a', 'b', 2],
     ['a', 1, 2, 'b'], ['a', 1, 'b', 2], ['a', 'b', 1, 2]]
    """
    def treeCrawl(path, i, j):
        if i == len(list1) and j == len(list2):
            yield path
            return
        if i < len(list1):
            yield from treeCrawl(path + [list1[i]], i + 1, j)
        if j < len(list2):
            yield from treeCrawl(path + [list2[j]], i, j + 1)

    yield from treeCrawl([], 0, 0)




############## for tensor caculus
def permuteTupleEntries(arg1, arg2):
    """
    Apply a permutation to the entries of a tuple or list.

    This function takes a tuple or list *arg1*, containing integers in the range
    [0, ..., k-1] for some integer k, and applies a permutation *arg2* to the
    entries of *arg1*. The result is returned as a new tuple.

    Parameters
    ----------
    arg1 : tuple or list
        A tuple or list containing integers in the range [0, ..., k-1].
    arg2 : list
        A list representing a permutation of [0, 1, ..., k-1].

    Returns
    -------
    tuple
        A new tuple with the entries of *arg1* permuted according to *arg2*.

    Examples
    --------
    >>> permuteTupleEntries((1, 2, 0), [2, 0, 1])
    (0, 1, 2)

    Notes
    -----
    - The elements of *arg1* must be valid indices in the permutation *arg2*.
    - If *arg1* contains out-of-range values, the function will raise an error.

    Raises
    ------
    ValueError
        If the values in *arg1* are out of the range defined by the length of *arg2*.
    """
    if not all(0 <= x < len(arg2) for x in arg1):
        raise ValueError(
            "Entries of arg1 must be in the range of [0, ..., len(arg2)-1]."
        )

    return tuple(arg2[arg1[j]] for j in range(len(arg1)))

def permuteTuple(arg1, arg2):
    """
    Apply a permutation to the order of a tuple or list.

    This function takes a tuple or list *arg1* of length *k*, and applies the
    permutation *arg2* (a permutation of [0, 1, ..., k-1]) to reorder its
    elements. The result is returned as a new tuple.

    Parameters
    ----------
    arg1 : tuple or list
        A tuple or list of length *k* whose elements will be permuted.
    arg2 : list
        A list representing a permutation of [0, 1, ..., k-1].

    Returns
    -------
    tuple
        A new tuple with the elements of *arg1* rearranged according to *arg2*.

    Examples
    --------
    >>> permuteTuple((1, 2, 3), [2, 0, 1])
    (3, 1, 2)

    Notes
    -----
    - The length of *arg2* must match the length of *arg1*.
    - If *arg2* contains invalid indices, the function will raise an error.

    Raises
    ------
    ValueError
        If the length of *arg2* does not match the length of *arg1*, or if *arg2*
        contains invalid indices.
    """
    if len(arg1) != len(arg2):
        raise ValueError("The length of arg2 must match the length of arg1.")
    if not all(0 <= x < len(arg1) for x in arg2):
        raise ValueError("arg2 must contain valid indices for arg1.")

    return tuple(arg1[j] for j in arg2)

def permuteArray(arg1, arg2):
    """
    Permute the indices of a k-dimensional array representing a multilinear operator.

    This function takes a k-dimensional array *arg1* that represents a multilinear operator
    on n-dimensional space, and applies the permutation *arg2* (a permutation of the
    coordinate indices [0, 1, ..., k-1]) to the index coordinate tuples of the array. The
    result is a new array with permuted indices.

    Parameters
    ----------
    arg1 : array-like
        A k-dimensional array representing a multilinear operator.
    arg2 : list
        A permutation of the coordinate indices [0, 1, ..., k-1].

    Returns
    -------
    MutableDenseNDimArray
        A new k-dimensional array with permuted indices.

    Examples
    --------
    >>> from sympy import MutableDenseNDimArray
    >>> A = MutableDenseNDimArray.zeros(2, 2, 2)
    >>> A[0, 1, 0] = 5
    >>> A[1, 0, 1] = 7
    >>> permuteArray(A, [2, 0, 1])
    MutableDenseNDimArray([[[0, 5], [0, 0]], [[0, 0], [7, 0]]])

    Notes
    -----
    - The length of *arg2* must match the number of dimensions of *arg1*.
    - The permutation is applied to the indices of the array, not its values.

    Raises
    ------
    ValueError
        If *arg2* is not a valid permutation of the indices.
    """
    if len(arg2) != len(arg1.shape):
        raise ValueError(
            "The length of arg2 must match the number of dimensions of arg1."
        )

    # Create a new array to store permuted values
    newArray = MutableDenseNDimArray.zeros(*arg1.shape)

    # Process the generator returned by chooseOp lazily
    for iListLoc in chooseOp(range(arg1.shape[0]), len(arg1.shape)):
        newListLoc = permuteTuple(iListLoc, arg2)
        newArray[newListLoc] = arg1[iListLoc]

    return newArray

def alternatingPartOfArray(arg1):
    """
    Calculate the alternating part of a multilinear operator.

    This function computes the alternating part of a k-dimensional array
    representing a multilinear operator. The alternating part is calculated by
    summing over all possible permutations of the array’s indices, applying
    the permutation sign (even or odd).

    Parameters
    ----------
    arg1 : Array-like
        A k-dimensional array representing a multilinear operator. The
        dimensions of the array must be equal (e.g., a square array).

    Returns
    -------
    MutableDenseNDimArray
        The alternating part of the input array.

    Examples
    --------
    >>> from sympy import MutableDenseNDimArray
    >>> A = MutableDenseNDimArray.zeros(3, 3, 3)
    >>> A[0, 1, 2] = 5
    >>> alternatingPartOfArray(A)
    MutableDenseNDimArray([[[0, 0, 0], [0, 0, 5], [0, -5, 0]], [[0, 0, -5], [0, 0, 0], [5, 0, 0]], [[0, 5, 0], [-5, 0, 0], [0, 0, 0]]])

    Notes
    -----
    - The input array must have equal dimensions.
    - The result is obtained by summing over all possible permutations of
      the indices, weighted by the permutation sign.

    Raises
    ------
    ValueError
        If the input array is not square or has unequal dimensions.
    """
    if isinstance(arg1, MutableDenseNDimArray):
        if len(set(arg1.shape)) == 1:
            permListLoc = chooseOp(
                range(len(arg1.shape)),
                len(arg1.shape),
                withoutReplacement=True,
                withOrder=True,
            )
            resultArray = MutableDenseNDimArray.zeros(*arg1.shape)

            # Process permutations lazily from the generator
            for perm in permListLoc:
                resultArray += permSign(perm) * permuteArray(arg1, perm)

            return resultArray

    raise ValueError("Input array must have equal dimensions.")



import sympy as sp

from .eds import abstDFAtom, abstDFMonom, abstract_DF, abstract_ZF, zeroFormAtom


class DF_representation(sp.Basic):
    def __new__(cls, row_count=None, col_count=None, array_data=None):
        if all(arg is None for arg in [col_count,array_data]):
            if row_count is None:
                array_data = tuple()
            else:
                array_data = row_count  # for optional syntax where only array data is given


        dgcv_classes = [zeroFormAtom, abstDFAtom, abstDFMonom, abstract_DF, abstract_ZF]
        supported_classes = tuple(dgcv_classes + [sp.Expr, sp.NumberSymbol, float, int])

        if callable(array_data):
            if row_count is None or col_count is None:
                raise ValueError("When using a lambda function for array_data, row_count and col_count must be specified.")
            array_data = tuple(tuple(array_data(j, k) for k in range(col_count)) for j in range(row_count))

        if row_count is None or col_count is None:
            if isinstance(array_data, dict):
                if array_data:
                    row_count = max(key[0] for key in array_data.keys()) + 1
                    col_count = max(key[1] for key in array_data.keys()) + 1
                else:
                    row_count, col_count = 0, 0
            elif isinstance(array_data, (list, tuple)):
                if all(isinstance(entry, (list, tuple)) for entry in array_data):
                    row_count = len(array_data)
                    col_count = max(len(row) for row in array_data) if row_count > 0 else 0
                else:
                    row_count = 1
                    col_count = len(array_data)
                    array_data = (tuple(array_data),)

        def validate_keys(key, r_count, c_count):
            if not (isinstance(key, tuple) and len(key) == 2):
                raise TypeError('Keys must be length-2 integers within specified row and column bounds')
            if any(entry < 0 for entry in key) or c_count <= key[1] or r_count <= key[0]:
                raise TypeError('Keys must be within row and column bounds')
            return True

        sparse_data = tuple()
        if isinstance(array_data, dict) and all(validate_keys(key, row_count, col_count) for key in array_data.keys()):
            if not all(isinstance(value, supported_classes) for value in array_data.values()):
                raise ValueError(f'DF_representation only supports entries from: {", ".join(str(cls) for cls in supported_classes)}')
            sparse_data = tuple({k: v for k, v in array_data.items() if v != 0}.items())
            array_data = tuple(tuple(array_data.get((j, k), 0) for k in range(col_count)) for j in range(row_count))
        elif isinstance(array_data, (list, tuple)) and all(isinstance(entry, (list, tuple)) for entry in array_data):
            if len(array_data) != row_count:
                raise ValueError('Incompatible row count')
            for row in array_data:
                if len(row) != col_count:
                    raise ValueError('Incompatible column count')
                if not all(isinstance(entry, supported_classes) for entry in row):
                    raise ValueError(f'Only supports entries from: {", ".join(str(cls) for cls in supported_classes)}')

            array_data = tuple(tuple(row) for row in array_data)
            sparse_data = {array_data[j][k] for j in range(row_count) for k in range(col_count)}

        obj = sp.Basic.__new__(cls, array_data)
        obj.row_count = row_count
        obj.col_count = col_count
        obj.array = array_data
        obj.sparse_data = sparse_data
        return obj

    def __init__(self, row_count=0, col_count=0, array_data=tuple()):
        pass

    def __getitem__(self, key):
        if isinstance(key, tuple):
            row_key, col_key = key
        else:
            return self.array[key]  # Allow row-wise access

        if isinstance(row_key, int) and isinstance(col_key, int):
            return self.array[row_key][col_key]
        else:
            new_data = [row[col_key] for row in self.array[row_key]]
            return DF_representation(len(new_data), len(new_data[0]) if new_data else 0, new_data)

    def __iter__(self):
        for row in self.array:
            for entry in row:
                yield entry

    def __eq__(self, other):
        if not isinstance(other, DF_representation):
            return False
        return self.array == other.array

    def __add__(self, other):
        if not isinstance(other, DF_representation):
            raise TypeError("Can only add DF_representation objects")
        if self.row_count != other.row_count or self.col_count != other.col_count:
            raise ValueError("Matrix dimensions must match for addition")
        new_data = [[self.array[i][j] + other.array[i][j] for j in range(self.col_count)] for i in range(self.row_count)]
        return DF_representation(self.row_count, self.col_count, new_data)

    def __sub__(self, other):
        if not isinstance(other, DF_representation):
            raise TypeError("Can only subtract DF_representation objects")
        if self.row_count != other.row_count or self.col_count != other.col_count:
            raise ValueError("Matrix dimensions must match for subtraction")
        new_data = [[self.array[i][j] - other.array[i][j] for j in range(self.col_count)] for i in range(self.row_count)]
        return DF_representation(self.row_count, self.col_count, new_data)

    def __mul__(self, scalar):
        if not isinstance(scalar, (abstract_ZF,zeroFormAtom,sp.Expr, sp.NumberSymbol, float, int)):
            raise TypeError("Can only multiply by scalar values")
        new_data = [[scalar * self.array[i][j] for j in range(self.col_count)] for i in range(self.row_count)]
        return DF_representation(self.row_count, self.col_count, new_data)

    __rmul__ = __mul__

    def __matmul__(self, other):
        if not isinstance(other, DF_representation):
            raise TypeError("Can only multiply with another DF_representation")
        if self.col_count != other.row_count:
            raise ValueError("Inner matrix dimensions must match for multiplication")
        new_data = [[sum(self.array[i][k] * other.array[k][j] for k in range(self.col_count)) for j in range(other.col_count)] for i in range(self.row_count)]
        return DF_representation(self.row_count, other.col_count, new_data)

    def __repr__(self):
        return f"DF_representation(row_count={self.row_count}, col_count={self.col_count}, array={self.array})"

    def __str__(self):
        return '\n'.join([' '.join(map(str, row)) for row in self.array])

    def flatten(self):
        return [entry for entry in self]

    def _latex(self,printer=None):
        rows = []
        for row in self.array:
            rendered_row = []
            for entry in row:
                if hasattr(entry, "_latex"):
                    latex = entry._latex()
                else:
                    latex = str(entry)
                rendered_row.append(latex)
            rows.append(" & ".join(rendered_row))
        return r"\begin{pmatrix}" + " \\\\ ".join(rows) + r"\end{pmatrix}"

    def _repr_latex_(self):
        rows = []
        for row in self.array:
            rendered_row = []
            for entry in row:
                if hasattr(entry, "_latex"):
                    latex = entry._latex()
                else:
                    latex = str(entry)
                rendered_row.append(latex)
            rows.append(" & ".join(rendered_row))
        latex_str = r"\begin{pmatrix}" + " \\\\ ".join(rows) + r"\end{pmatrix}"
        return "$" + latex_str + "$"

    def applyfunc(self, func):
        new_data = [
            [func(self.array[i][j]) for j in range(self.col_count)]
            for i in range(self.row_count)
        ]
        return DF_representation(self.row_count, self.col_count, new_data)

    def _dgcv_eds_applyfunc(self, func):
        return self.applyfunc(func)

    def _eval_conjugate(self):
        def _custom_conj(expr):
            if isinstance(expr,(abstract_ZF,zeroFormAtom,abstDFAtom,abstDFMonom,abstract_DF)):
                return expr._eval_conjugate()
            else:
                return sp.conjugate(expr)
        return self.applyfunc(_custom_conj)

    def _eval_simplify(self, ratio=None, measure=None, inverse=True, doit=True, rational=True, expand=False, **kwargs):
        return self.applyfunc(lambda entry: sp.simplify(entry,ratio=ratio, measure=measure, inverse=inverse, doit=doit, rational=rational, expand=expand, **kwargs))

    def _eval_canonicalize(self,depth = 1000):
        def _canon(obj):
            if hasattr(obj,'_eval_canonicalize'):
                return obj._eval_canonicalize(depth=depth)
            return obj
        return self.applyfunc(_canon)


    def _subs_DGCV(self, data, with_diff_corollaries=False):
        # an alias for regular subs so that other functions can know the with_diff_corollaries keyword is available
        return self.subs(data, with_diff_corollaries = with_diff_corollaries)

    def subs(self,subs_data,with_diff_corollaries=False):
        def _custom_subs(expr):
            if hasattr(expr,'_subs_DGCV'):
                return expr._subs_DGCV(subs_data,with_diff_corollaries=with_diff_corollaries)
            if hasattr(expr,'subs'):
                return expr.subs(subs_data)
            else:
                return expr
        return self.applyfunc(_custom_subs)




"""
_display.py

This module provides functions for customizing the display in Jupyter notebooks.
The `load_fonts()` function loads web fonts and applies them to the notebook
environment using HTML and CSS.

Functions
---------
- load_fonts: Injects HTML and CSS to load Google Fonts and apply custom fonts to the notebook.
"""


############# for printing
import sympy as sp
from IPython.display import HTML, Latex, display
from sympy.printing.latex import LatexPrinter

############# DGCV classes to format
from .dgcv_core import (
    DFClass,
    DGCVPolyClass,
    STFClass,
    VFClass,
    symToHol,
    symToReal,
    tensorField,
)
from .filtered_structures import Tanaka_symbol
from .finite_dim_algebras import AlgebraElement, FAClass
from .Riemannian_geometry import metricClass


def LaTeX(obj, removeBARs=False):
    """
    Custom LaTeX function for DGCV. Extends sympy.latex() to support application to DGCV classes.

    Parameters
    ----------
    obj : any
        The object to convert to LaTeX. Can be a SymPy expression, DGCV object, or a list/tuple of such objects.

    Returns
    -------
    str
        The LaTeX-formatted string.
    """

    def filter(term):
        if removeBARs:
            return sp.latex(term)
        if isinstance(term, (DFClass, VFClass, STFClass,tensorField, DGCVPolyClass)):
            if term._varSpace_type == "real":
                return sp.latex(symToReal(term))
            elif term._varSpace_type == "complex":
                return sp.latex(symToHol(term))
            else:
                return sp.latex(term)
        elif isinstance(term, FAClass):
            return _alglabeldisplayclass(term.label)._repr_latex_()
        elif isinstance(term, AlgebraElement):
            return _alglabeldisplayclass(term.algebra.label, term)._repr_latex_()
        elif isinstance(term, Tanaka_symbol):
            return "Tanaka_symbol Class"
        else:
            return sp.latex(symToHol(term))

    def strip_dollar_signs(latex_str):
        """Strip leading and trailing $ or $$ signs from a LaTeX string."""
        # if latex_str == None:
        #     return ''
        if latex_str is None:
            return latex_str
        if latex_str.startswith("$$") and latex_str.endswith("$$"):
            return latex_str[2:-2]
        if latex_str.startswith("$") and latex_str.endswith("$"):
            return latex_str[1:-1]
        return latex_str

    if isinstance(obj, list):
        latex_elements = [strip_dollar_signs(filter(elem)) for elem in obj]
        return r"\left[ " + ", ".join(latex_elements) + r" \right]"
    elif isinstance(obj, tuple):
        latex_elements = [strip_dollar_signs(filter(elem)) for elem in obj]
        return r"\left( " + ", ".join(latex_elements) + r" \right)"
    elif isinstance(obj, set):
        latex_elements = [strip_dollar_signs(filter(elem)) for elem in obj]
        return r"\left\{ " + ", ".join(latex_elements) + r" \right\}"

    else:
        # Apply sympy.latex() for non-list/tuple objects
        return strip_dollar_signs(filter(obj))


def display_DGCV(*args):
    for j in args:
        _display_DGCV_single(j)


def _display_DGCV_single(arg):
    if isinstance(arg, str):
        display(Latex(arg))
    elif isinstance(
        arg,
        (sp.Expr, metricClass, DFClass, VFClass, STFClass,tensorField, DGCVPolyClass,Tanaka_symbol),
    ):
        _complexDisplay(arg)
    elif isinstance(arg, FAClass):
        _complexDisplay(_alglabeldisplayclass(arg.label))
    elif isinstance(arg, AlgebraElement):
        _complexDisplay(_alglabeldisplayclass(arg.algebra.label, ae=arg))
    else:
        display(arg)


def _complexDisplay(*args):
    """
    Taking DGCV expressions in *args* written in terms of symbolic conjugate variables, displays them with actual complex conjugates
    """
    display(*[symToHol(j, simplify_everything=False) for j in args])


class _alglabeldisplayclass(sp.Basic):

    def __new__(cls, label, ae=None):
        obj = sp.Basic.__new__(cls, label)
        return obj

    def __init__(self, label, ae=None):
        self.label = str(label)
        self.ae = ae

    @staticmethod
    def format_algebra_label(label):
        r"""Wrap the algebra label in \mathfrak{} if all characters are lowercase, and subscript any numeric suffix."""
        if label[-1].isdigit():
            # Split into text and number parts for subscript formatting
            label_text = "".join(filter(str.isalpha, label))
            label_number = "".join(filter(str.isdigit, label))
            if label_text.islower():
                return rf"\mathfrak{{{label_text}}}_{{{label_number}}}"
            return rf"{label_text}_{{{label_number}}}"
        elif label.islower():
            return rf"\mathfrak{{{label}}}"
        return label

    @staticmethod
    def format_ae(ae):
        if ae is not None:
            terms = []
            for coeff, basis_label in zip(ae.coeffs, ae.algebra.basis_labels):
                if coeff == 0:
                    continue  # Skip zero terms
                elif coeff == 1:
                    terms.append(rf"{basis_label}")  # Suppress 1 as coefficient
                elif coeff == -1:
                    terms.append(
                        rf"-{basis_label}"
                    )  # Suppress 1 but keep the negative sign
                else:
                    # Check if the coefficient has more than one term (e.g., 1 + I)
                    if isinstance(coeff, sp.Expr) and len(coeff.args) > 1:
                        terms.append(
                            rf"({sp.latex(coeff)}) \cdot {basis_label}"
                        )  # Wrap multi-term coefficient in parentheses
                    else:
                        terms.append(
                            rf"{sp.latex(coeff)} \cdot {basis_label}"
                        )  # Single-term coefficient

            # Handle special case: all zero coefficients
            if not terms:
                return rf"$0 \cdot {ae.algebra.basis_labels[0]}$"

            # Join terms with proper LaTeX sign handling
            result = " + ".join(terms).replace("+ -", "- ")
            return rf"{result}"

    def _repr_latex_(self):
        if self.ae:
            return _alglabeldisplayclass.format_algebra_label(self.label)
        else:
            return _alglabeldisplayclass.format_ae(self.ae)

    def __str__(self):
        return self.label


def load_fonts():
    font_links = """
    <link href="https://fonts.googleapis.com/css2?family=Orbitron:wght@400;700&family=Press+Start+2P&family=Roboto:wght@400;700&display=swap" rel="stylesheet">
    <style>
    body {
        font-family: 'Roboto', sans-serif;
    }
    </style>
    """
    display(HTML(font_links))



# DGCV-specific SymPy LatexPrinter for VFClass and DFClass
class DGCVLatexPrinter(LatexPrinter):
    def _print_VFClass(self, expr):
        return expr._repr_latex_()

    def _print_DFClass(self, expr):
        return expr._repr_latex_()


def DGCV_collection_latex_printer(obj):
    if isinstance(obj, (tuple, list)):
        return tuple(
            Latex(element._repr_latex_() if hasattr(element, "_repr_latex_") else sp.latex(element))
            for element in obj
        )
    return None


# def DGCV_latex_printer(obj, **kwargs):
#     if isinstance(
#         obj,
#         (
#             VFClass,
#             DFClass,
#             STFClass,
#             tensorField,
#             metricClass,
#             FAClass,
#             AlgebraElement,
#             DGCVPolyClass,
#             Tanaka_symbol
#         ),
#     ):
#         latex_str = obj._repr_latex_()
#         return latex_str.strip("$")
#     elif isinstance(obj, (list, tuple)):
#         latex_elements = [DGCV_latex_printer(elem) for elem in obj]
#         return r"\left( " + r" , ".join(latex_elements) + r" \right)"
#     return latex(obj, **kwargs)

def DGCV_latex_printer(obj, **kwargs):
    if obj is None:
        return ''
    if LaTeX(obj) is None:
        return ''
    return LaTeX(obj).strip("$")

def DGCV_init_printing(*args, **kwargs):
    from sympy import init_printing

    kwargs["latex_printer"] = DGCV_latex_printer
    init_printing(*args, **kwargs)



import warnings

import sympy as sp

from ._safeguards import (
    create_key,
    retrieve_passkey,
    validate_label,
    validate_label_list,
)
from ._tensor_field_printers import (
    tensor_latex_helper,
    tensor_VS_printer,
)
from .combinatorics import shufflings
from .config import _cached_caller_globals, get_variable_registry
from .dgcv_core import clearVar, listVar


# vector space class
class vectorSpace(sp.Basic):
    def __new__(cls, dimension, *args, **kwargs):
        if not isinstance(dimension, int) or dimension < 0:
            raise TypeError("vectorSpace expected dimension to be a positive int.")

        # Create the new instance
        obj = sp.Basic.__new__(cls, dimension)
        return obj

    def __init__(
        self,
        dimension,
        grading=None,
        _label=None,
        _basis_labels=None,
        _calledFromCreator=None,
    ):
        self.dimension = dimension

        # Detect if initialized from creator (using the passkey)
        if _calledFromCreator == retrieve_passkey():
            self.label = _label
            self.basis_labels = tuple(_basis_labels) if _basis_labels else None
            self._registered = True
        else:
            self.label = "Alg_" + create_key()  # Assign a random label
            self.basis_labels = None
            self._registered = False

        def validate_and_adjust_grading_vector(vector, dimension):
            """
            Validates and adjusts a grading vector to match the vector space's dimension.

            Parameters
            ----------
            vector : list, tuple, or sympy.Tuple
                The grading vector to validate and adjust.
            dimension : int
                The dimension of the vector space.

            Returns
            -------
            sympy.Tuple
                The validated and adjusted grading vector.
            """
            if not isinstance(vector, (list, tuple, sp.Tuple)):
                raise ValueError(
                    "Grading vector must be a list, tuple, or SymPy Tuple."
                )
            vector = list(vector)

            if len(vector) < dimension:
                warnings.warn(
                    f"Grading vector is shorter than the dimension ({len(vector)} < {dimension}). "
                    f"Padding with zeros to match the dimension.",
                    UserWarning,
                )
                vector += [0] * (dimension - len(vector))
            elif len(vector) > dimension:
                warnings.warn(
                    f"Grading vector is longer than the dimension ({len(vector)} > {dimension}). "
                    f"Truncating to match the dimension.",
                    UserWarning,
                )
                vector = vector[:dimension]

            # Validate components
            for i, component in enumerate(vector):
                if not isinstance(component, (int, float, sp.Basic)):
                    raise ValueError(
                        f"Invalid component in grading vector at index {i}: {component}. "
                        f"Expected int, float, or sympy.Expr."
                    )

            return sp.Tuple(*vector)

        # Process grading
        if grading is None:
            # Default to a single grading vector [0, 0, ..., 0]
            self.grading = (tuple([0] * self.dimension),)
        else:
            # Handle single or multiple grading vectors
            if isinstance(grading, (list, tuple)) and all(
                isinstance(g, (list, tuple, sp.Tuple)) for g in grading
            ):
                # Multiple grading vectors provided
                self.grading = tuple(
                    validate_and_adjust_grading_vector(vector, self.dimension)
                    for vector in grading
                )
            else:
                # Single grading vector provided
                self.grading = (
                    validate_and_adjust_grading_vector(grading, self.dimension),
                )

        # Set the number of grading vectors
        self._gradingNumber = len(self.grading)

        # Initialize basis
        self.basis = tuple(
            vectorSpaceElement(
                self,
                [1 if i == j else 0 for j in range(self.dimension)],
                1
            )
            for i in range(self.dimension)
        )

    def __eq__(self, other):
        if not isinstance(other, vectorSpace):
            return NotImplemented
        return (
            self.dimension == other.dimension and
            self.label == other.label and
            self.basis_labels == other.basis_labels and
            self.grading == other.grading and
            self.basis == other.basis
        )

    def __hash__(self):
        return hash((self.dimension, self.label, self.basis_labels, self.grading, self.basis))

    def __iter__(self):
        return iter(self.basis) 

    def __contains__(self, item):
        return item in self.basis

    def __repr__(self):
        """
        Provides a detailed representation of the vectorSpace object.
        Raises a warning if the instance is unregistered.
        """
        if not self._registered:
            warnings.warn(
                "This vectorSpace instance was initialized without an assigned label. "
                "It is recommended to initialize vectorSpace objects with DGCV creator functions like `createVectorSpace` instead.",
                UserWarning,
            )
        return (
            f"vectorSpace(dim={self.dimension}, grading={self.grading}, "
            f"label={self.label}, basis_labels={self.basis_labels}"
        )

    def __str__(self):
        """
        Provides a string representation of the vectorSpace object.
        Raises a warning if the instance is unregistered.
        """
        if not self._registered:
            warnings.warn(
                "This vectorSpace instance was initialized without an assigned label. "
                "It is recommended to initialize vectorSpace objects with DGCV creator functions like `createVectorSpace` instead.",
                UserWarning,
            )

        def format_basis_label(label):
            return label

        formatted_label = self.label if self.label else "Unnamed VS"
        formatted_basis_labels = (
            ", ".join([format_basis_label(bl) for bl in self.basis_labels])
            if self.basis_labels
            else "No basis labels assigned"
        )
        return (
            f"Vector Space: {formatted_label}\n"
            f"Dimension: {self.dimension}\n"
            f"Grading: {self.grading}\n"
            f"Basis: {formatted_basis_labels}"
        )

    def _display_DGCV_hook(self):
        """
        Hook for DGCV-specific display customization.
        Raises a warning if the instance is unregistered.
        """
        if not self._registered:
            warnings.warn(
                "This vectorSpace instance was initialized without an assigned label. "
                "It is recommended to initialize vectorSpace objects with DGCV creator functions like `createVectorSpace` instead.",
                UserWarning,
            )

        def format_VS_label(label):
            r"""Wrap the vector space label in \mathfrak{} if all characters are lowercase, and subscript any numeric suffix."""
            if label and label[-1].isdigit():
                label_text = "".join(filter(str.isalpha, label))
                label_number = "".join(filter(str.isdigit, label))
                if label_text.islower():
                    return rf"\mathfrak{{{label_text}}}_{{{label_number}}}"
                return rf"{label_text}_{{{label_number}}}"
            elif label and label.islower():
                return rf"\mathfrak{{{label}}}"
            return label or "Unnamed Vector Space"

        return format_VS_label(self.label)

    def _repr_latex_(self):
        """
        Provides a LaTeX representation of the vectorSpace object for Jupyter notebooks.
        Raises a warning if the instance is unregistered.
        """
        if not self._registered:
            warnings.warn(
                "This vectorSpace instance was initialized without an assigned label. "
                "It is recommended to initialize vectorSpace objects with DGCV creator functions like `createVectorSpace` instead.",
                UserWarning,
            )

        def format_VS_label(label):
            r"""
            Formats a vector space label for LaTeX. Handles:
            1. Labels with an underscore, splitting into two parts:
            - The first part goes into \mathfrak{} if it is lowercase.
            - The second part becomes a LaTeX subscript.
            2. Labels without an underscore:
            - Checks if the label ends in a numeric tail for subscripting.
            - Otherwise wraps the label in \mathfrak{} if it is entirely lowercase.

            Parameters
            ----------
            label : str
                The vector space label to format.

            Returns
            -------
            str
                A LaTeX-formatted vector space label.
            """
            if not label:
                return "Unnamed Vector Space"

            if "_" in label:
                # Split the label at the first underscore
                main_part, subscript_part = label.split("_", 1)
                if main_part.islower():
                    return rf"\mathfrak{{{main_part}}}_{{{subscript_part}}}"
                return rf"{main_part}_{{{subscript_part}}}"

            if label[-1].isdigit():
                # Split into text and numeric parts for subscripting
                label_text = "".join(filter(str.isalpha, label))
                label_number = "".join(filter(str.isdigit, label))
                if label_text.islower():
                    return rf"\mathfrak{{{label_text}}}_{{{label_number}}}"
                return rf"{label_text}_{{{label_number}}}"

            if label.islower():
                # Wrap entirely lowercase labels in \mathfrak{}
                return rf"\mathfrak{{{label}}}"

            # Return the label as-is if no special conditions apply
            return label

        def format_basis_label(label):
            return rf"{label}" if label else "e_i"

        formatted_label = format_VS_label(self.label)
        formatted_basis_labels = (
            ", ".join([format_basis_label(bl) for bl in self.basis_labels])
            if self.basis_labels
            else "No basis labels assigned"
        )
        return (
            f"Vector Space: ${formatted_label}$, Basis: ${formatted_basis_labels}$, "
            f"Dimension: ${self.dimension}$, Grading: ${sp.latex(self.grading)}$"
        )

    def _sympystr(self):
        """
        SymPy string representation for vectorSpace.
        Raises a warning if the instance is unregistered.
        """
        if not self._registered:
            warnings.warn(
                "This vectorSpace instance was initialized without an assigned label. "
                "It is recommended to initialize vectorSpace objects with DGCV creator functions like `createVectorSpace` instead.",
                UserWarning,
            )

        if self.label:
            return f"vectorSpace({self.label}, dim={self.dimension})"
        else:
            return f"vectorSpace(dim={self.dimension})"

    def subspace_basis(self, elements):
        """
        Computes a basis of subspace spanned by given set of elements.

        Parameters
        ----------
        elements : list
            A list of vectorSpaceElement instances.

        Returns
        -------
        list of vectorSpaceElement
            basis of subspace
        """

        if not all(isinstance(j,vectorSpaceElement) for j in elements) or not all(j.vectorSpace==self for j in elements) or len(set([j.valence for j in elements]))!=1:
            raise TypeError('vectorSpace.subspace_basis expects a list of elements from the calling vectorSpace instance.')

        # Perform linear independence check
        span_matrix = sp.Matrix.hstack(*[el.coeffs for el in elements])
        linearly_independent = span_matrix.rank() == len(elements)

        if linearly_independent:
            return elements

        rref_matrix, pivot_columns = span_matrix.rref()

        # Extract the linearly independent basis
        return [vectorSpaceElement(self,elements[i],) for i in pivot_columns]

# vector space element class
class vectorSpaceElement(sp.Basic):
    def __new__(cls, VS, coeffs, valence):
        if not isinstance(VS, vectorSpace):
            raise TypeError(
                "vectorSpaceElement expects the first argument to be an instance of vectorSpace."
            )
        if valence not in {0, 1}:
            raise TypeError("vectorSpaceElement expects third argument to be 0 or 1.")

        coeffs = tuple(coeffs)

        obj = sp.Basic.__new__(cls, VS, coeffs, valence)
        return obj

    def __init__(self, VS, coeffs, valence):
        self.vectorSpace = VS
        self.coeffs = tuple(coeffs)  # Store coeffs as a tuple instead of a Matrix
        self.valence = valence

    def __eq__(self, other):
        if not isinstance(other, vectorSpaceElement):
            return NotImplemented
        return (
            self.vectorSpace == other.vectorSpace and
            self.coeffs == other.coeffs and
            self.valence == other.valence
        )

    def __hash__(self):
        return hash((self.vectorSpace, self.coeffs, self.valence))

    def __str__(self):
        """
        Custom string representation for vectorSpaceElement.
        Displays the linear combination of basis elements with coefficients.
        Handles unregistered parent vector space by raising a warning.
        """
        if not self.vectorSpace._registered:
            warnings.warn(
                "This vectorSpaceElement's parent vector space (vectorSpace) was initialized without an assigned label. "
                "It is recommended to initialize vectorSpace objects with DGCV creator functions like `createVectorSpace` instead.",
                UserWarning,
            )

        terms = []
        for coeff, basis_label in zip(
            self.coeffs,
            self.vectorSpace.basis_labels
            or [f"e_{i+1}" for i in range(self.vectorSpace.dimension)],
        ):
            if coeff == 0:
                continue
            elif coeff == 1:
                if self.valence==1:
                    terms.append(f"{basis_label}")
                else:
                    terms.append(f"{basis_label}^\'\'")
            elif coeff == -1:
                if self.valence==1:
                    terms.append(f"-{basis_label}")
                else:
                    terms.append(f"-{basis_label}^\'\'")
            else:
                if isinstance(coeff, sp.Expr) and len(coeff.args) > 1:
                    if self.valence==1:
                        terms.append(f"({coeff}) * {basis_label}")
                    else:
                        terms.append(f"({coeff}) * {basis_label}^\'\'")
                else:
                    if self.valence==1:
                        terms.append(f"{coeff} * {basis_label}")
                    else:
                        terms.append(f"{coeff} * {basis_label}^\'\'")

        if not terms:
            if self.valence==1:
                return f"0 * {self.vectorSpace.basis_labels[0] if self.vectorSpace.basis_labels else 'e_1'}"
            else:
                return f"0 * {self.vectorSpace.basis_labels[0] if self.vectorSpace.basis_labels else 'e_1'}^\'\'"

        return " + ".join(terms).replace("+ -", "- ")

    def _repr_latex_(self):
        """
        Provides a LaTeX representation of vectorSpaceElement for Jupyter notebooks.
        Handles unregistered parent vector space by raising a warning.
        """
        if not self.vectorSpace._registered:
            warnings.warn(
                "This vectorSpaceElement's parent vector space (vectorSpace) was initialized without an assigned label. "
                "It is recommended to initialize vectorSpace objects with DGCV creator functions like `createVectorSpace` instead.",
                UserWarning,
            )

        terms = []
        for coeff, basis_label in zip(
            self.coeffs,
            self.vectorSpace.basis_labels
            or [f"e_{i+1}" for i in range(self.vectorSpace.dimension)],
        ):
            if coeff == 0:
                continue
            elif coeff == 1:
                if self.valence==1:
                    terms.append(rf"{basis_label}")
                else:
                    terms.append(rf"{basis_label}^*")
            elif coeff == -1:
                if self.valence==1:
                    terms.append(rf"-{basis_label}")
                else:
                    terms.append(rf"-{basis_label}^*")
            else:
                if isinstance(coeff, sp.Expr) and len(coeff.args) > 1:
                    if self.valence==1:
                        terms.append(rf"({sp.latex(coeff)}) \cdot {basis_label}")
                    else:
                        terms.append(rf"({sp.latex(coeff)}) \cdot {basis_label}^*")
                else:
                    if self.valence==1:
                        terms.append(rf"{sp.latex(coeff)} \cdot {basis_label}")
                    else:
                        terms.append(rf"{sp.latex(coeff)} \cdot {basis_label}^*")

        if not terms:
            return rf"$0 \cdot {self.vectorSpace.basis_labels[0] if self.vectorSpace.basis_labels else 'e_1'}$"

        result = " + ".join(terms).replace("+ -", "- ")

        def format_VS_label(label):
            r"""
            Wrap the vector space label in \mathfrak{} if lowercase, and add subscripts for numeric suffixes or parts.
            """
            if "_" in label:
                main_part, subscript_part = label.split("_", 1)
                if main_part.islower():
                    return rf"\mathfrak{{{main_part}}}_{{{subscript_part}}}"
                return rf"{main_part}_{{{subscript_part}}}"
            elif label[-1].isdigit():
                label_text = "".join(filter(str.isalpha, label))
                label_number = "".join(filter(str.isdigit, label))
                if label_text.islower():
                    return rf"\mathfrak{{{label_text}}}_{{{label_number}}}"
                return rf"{label_text}_{{{label_number}}}"
            elif label.islower():
                return rf"\mathfrak{{{label}}}"
            return label

        return rf"$\text{{Element of }} {format_VS_label(self.vectorSpace.label)}: {result}$"

    def _sympystr(self):
        """
        SymPy string representation for vectorSpaceElement.
        Handles unregistered parent vector space by raising a warning.
        """
        if not self.vectorSpace._registered:
            warnings.warn(
                "This vectorSpaceElement's parent vector space (vectorSpace) was initialized without an assigned label. "
                "It is recommended to initialize vectorSpace objects with DGCV creator functions like `createVectorSpace` instead.",
                UserWarning,
            )

        coeffs_str = ", ".join(map(str, self.coeffs))
        if self.vectorSpace.label:
            return f"vectorSpaceElement({self.vectorSpace.label}, coeffs=[{coeffs_str}])"
        else:
            return f"vectorSpaceElement(coeffs=[{coeffs_str}])"

    def __repr__(self):
        """
        Representation of vectorSpaceElement.
        Shows the linear combination of basis elements with coefficients.
        Falls back to __str__ if basis_labels is None.
        """
        if self.vectorSpace.basis_labels is None:
            # Fallback to __str__ when basis_labels is None
            return str(self)

        terms = []
        for coeff, basis_label in zip(self.coeffs, self.vectorSpace.basis_labels):
            if coeff == 0:
                continue
            elif coeff == 1:
                if self.valence==1:
                    terms.append(f"{basis_label}")
                else:
                    terms.append(f"{basis_label}^\'\'")
            elif coeff == -1:
                if self.valence==1:
                    terms.append(f"-{basis_label}")
                else:
                    terms.append(f"-{basis_label}^\'\'")
            else:
                if isinstance(coeff, sp.Expr) and len(coeff.args) > 1:
                    if self.valence==1:
                        terms.append(f"({coeff}) * {basis_label}")
                    else:
                        terms.append(f"({coeff}) * {basis_label}^\'\'")
                else:
                    if self.valence==1:
                        terms.append(f"{coeff} * {basis_label}")
                    else:
                        terms.append(f"{coeff} * {basis_label}^\'\'")

        if not terms:
            if self.valence==1:
                return f"0*{self.vectorSpace.basis_labels[0]}"
            else:
                return f"0*{self.vectorSpace.basis_labels[0]}^\'\'"

        return " + ".join(terms).replace("+ -", "- ")

    def is_zero(self):
        for j in self.coeffs:
            if sp.simplify(j) != 0:
                return False
        else:
            return True

    def dual(self):
        return vectorSpaceElement(self.vectorSpace, self.coeffs, (self.valence+1)%2)

    def _convert_to_tp(self):
        return tensorProduct(self.vectorSpace,{(j,self.valence):self.coeffs[j] for j in range(self.vectorSpace.dimension)})

    def _recursion_contract_hom(self, other):
        return self._convert_to_tp()._recursion_contract_hom(other)

    def subs(self, subsData):
        newCoeffs = [sp.sympify(j).subs(subsData) for j in self.coeffs]
        return vectorSpaceElement(self.vectorSpace, newCoeffs)

    def __call__(self, other):
        if not isinstance(other, vectorSpaceElement) or other.vectorSpace!=self.vectorSpace or other.valence==self.valence:
            raise TypeError('`vectorSpaceElement.call()` can only be applied to `vectorSpaceElement` instances with the same vectorSpace but different valence attributes.')
        return sum([self.coeffs[j]*other.coeffs[j] for j in range(self.vectorSpace.dimension)])

    def __add__(self, other):
        if isinstance(other, vectorSpaceElement):
            if self.vectorSpace == other.vectorSpace and self.valence == other.valence:
                return vectorSpaceElement(
                    self.vectorSpace,
                    [self.coeffs[j] + other.coeffs[j] for j in range(len(self.coeffs))],
                    self.valence
                )
            else:
                raise TypeError(
                    "vectorSpaceElement operands for + must belong to the same vectorSpace."
                )
        else:
            raise TypeError(
                "Unsupported operand type(s) for + with the vectorSpaceElement class"
            )

    def __sub__(self, other):
        if isinstance(other, vectorSpaceElement):
            if self.vectorSpace == other.vectorSpace and self.valence == other.valence:
                return vectorSpaceElement(
                    self.vectorSpace,
                    [self.coeffs[j] - other.coeffs[j] for j in range(len(self.coeffs))],
                    self.valence
                )
            else:
                raise TypeError(
                    "vectorSpaceElement operands for - must belong to the same vectorSpace."
                )
        else:
            raise TypeError(
                "Unsupported operand type(s) for - with the vectorSpaceElement class"
            )

    def __mul__(self, other):
        """
        Multiplies two vectorSpaceElement objects by multiplying their coefficients
        and summing the results based on the vector space's structure constants. Also handles
        multiplication with scalars.

        Args:
            other (vectorSpaceElement) or (scalar): The vector space element or scalar to multiply with.

        Returns:
            vectorSpaceElement: The result of the multiplication.
        """
        if isinstance(other, (int, float, sp.Expr)):
            # Scalar multiplication case
            new_coeffs = [coeff * other for coeff in self.coeffs]
            return vectorSpaceElement(
                self.vectorSpace, new_coeffs, self.valence
            )
        else:
            raise TypeError(
                f"Multiplication is only supported for scalars, not {type(other)}"
            )

    def __rmul__(self, other):
        # If other is a scalar, treat it as commutative
        if isinstance(
            other, (int, float, sp.Expr)
        ):  # Handles numeric types and SymPy scalars
            return self * other  # Calls __mul__ (which is already implemented)
        else:
            raise TypeError(
                f"Right multiplication is only supported for scalars not {type(other)}"
            )

    def __matmul__(self, other):
        """Overload @ operator for tensor product."""
        if not isinstance(other, vectorSpaceElement) or other.vectorSpace!=self.vectorSpace:
            raise TypeError('`@` only supports tensor products between vectorSpaceElements instances with the same vectorSpace attribute')
        return tensorProduct(self.vectorSpace, {(j,k,self.valence,other.valence):self.coeffs[j]*other.coeffs[k] for j in range(self.vectorSpace.dimension) for k in range(self.vectorSpace.dimension)})

    def __xor__(self, other):
        if other == '':
            return self.dual()
            raise ValueError("Invalid operation. Use `^ ''` to denote the dual.")

class tensorProduct(sp.Basic):
    def __new__(cls, vector_space, coeff_dict):

        if not isinstance(coeff_dict, dict):
            raise ValueError("Coefficient dictionary must be a dictionary.")

        # Process the coefficient dictionary
        try:
            result = tensorProduct._process_coeffs_dict(coeff_dict)

            if not isinstance(result, tuple):
                raise TypeError(f"DEBUG: `tensorProduct._process_coeffs_dict` is expected to return a tuple, but produced type {type(result)}")

            if len(result) != 5:
                raise ValueError(f"DEBUG: `tensorProduct._process_coeffs_dict` is expected to return 5 values, but returned {len(result)}: {result}")

            processed_coeff_dict, max_degree, min_degree, prolongation_type, homogeneous_dicts = result

        except ValueError as ve:
            print(f"ValueError: {ve}\nDebug info: Check the return statement of _process_coeffs_dict.")

        except TypeError as te:
            print(f"TypeError: {te}\nDebug info: Make sure _process_coeffs_dict returns a tuple.")

        except Exception as e:
            print(f"Unexpected error: {e}\nDebug info: {result if 'result' in locals() else 'Function did not return any value.'}")

        obj = sp.Basic.__new__(cls, vector_space,  processed_coeff_dict)
        obj.vector_space = vector_space
        obj.coeff_dict = processed_coeff_dict
        obj.max_degree = max_degree
        obj.min_degree = min_degree
        obj.prolongation_type = prolongation_type
        obj.homogeneous_dicts = homogeneous_dicts
        obj.homogeneous = True if len(homogeneous_dicts)==1 else False
        return obj

    def __init__(self, vector_space, coeff_dict):
        self._weights = None
        self._leading_valence = None
        self._trailing_valence = None

    @staticmethod
    def _process_coeffs_dict(coeff_dict):
        """Process the coefficient dictionary."""
        if not coeff_dict:
            return {(0,) * 0: 0}, 0, 0, 0,[{(0,) * 0: 0}]

        max_degree = 0
        min_degree = -1
        valence = tuple()
        prolongation_type = None
        processed_dict = dict()
        homogeneous_dicts = dict()
        for key, value in coeff_dict.items():
            if not isinstance(key, tuple) or len(key)%2!=0 or not all(j in {0,1} for j in key[len(key)//2:]) or not all(isinstance(j,(int,float,sp.Expr)) for j in key[:len(key)//2]):
                raise ValueError("Keys in coeff_dict must be tuples of even length whose last half contain only 0s and 1s (indicating valence of the first half).")
            if 2*max_degree<len(key):
                max_degree = len(key)//2
                valence = key[len(key)//2:]
            if len(key)//2 in homogeneous_dicts:
                homogeneous_dicts[len(key)//2][key]=value
            else:
                homogeneous_dicts[len(key)//2]={key:value}
            if value != 0:
                processed_dict[key] = value
                if min_degree<0 or 2*min_degree>len(key):
                    min_degree = len(key)//2
                if prolongation_type!=-1 and all((key[len(key)//2]+j)%2==1 for j in key[len(key)//2+1:]):
                    if prolongation_type is None:
                        prolongation_type = key[len(key)//2]
                    elif prolongation_type!=key[len(key)//2]:
                        prolongation_type=-1
                else:
                    prolongation_type=-1
            if min_degree<0:
                min_degree=max_degree
        if prolongation_type is None:
            prolongation_type=-1

        # If all keys are removed (tensor is zero), restore a zero key
        if not processed_dict:
            processed_dict[((0,) * max_degree)+valence] = 0
        return processed_dict, max_degree, min_degree, prolongation_type, homogeneous_dicts

    @property
    def leading_valence(self):
        if self._leading_valence:
            return self._leading_valence
        lv = set()
        for k in self.coeff_dict:
            if len(k)>1:
                lv = lv|{k[len(k)//2]}
        if len(lv)==1:
            self._leading_valence = list(lv)[0]
        else:
            self._leading_valence = -1      # denoting exceptional cases
        return self._leading_valence

    @property
    def trailing_valence(self):
        if self._trailing_valence:
            return self._trailing_valence
        lv = set()
        for k in self.coeff_dict:
            if len(k)>0:
                lv = lv|{k[-1]}
        if len(lv)==1:
            self._trailing_valence = list(lv)[0]
        else:
            self._trailing_valence = -1     # denoting exceptional cases
        return self._trailing_valence

    @property
    def homogeneous_components(self):
        return [tensorProduct(self.vector_space,cd) for cd in self.homogeneous_dicts.values()]

    def _compute_weights(self):
        """Lazily compute weights for all terms in coeff_dict for each weight vector."""
        if not hasattr(self, "_weights"):
            self._weights = {}

        if not self._weights:  # Only compute if weights have not been cached
            self._weights = {}
            grading = self.vector_space.grading  # List of weight vectors

            for key, _ in self.coeff_dict.items():
                weight_list = []
                for weight_vector in grading:
                    weight = 0
                    for j, index in enumerate(key[:len(key)//2]):
                        weight += weight_vector[index] * (key[(len(key)//2)+j] * 2 - 1)
                    weight_list.append(weight)
                self._weights[key] = tuple(weight_list)

        return self._weights

    def compute_weight(self):
        weights = list(set((self._compute_weights()).values()))
        if len(weights)==1:
            return weights[0]
        else:
            return "NoW"

    def get_weighted_components(self, weight_list):
        """
        Return a new tensorProduct with components matching the given weight_list.

        Parameters:
        - weight_list: A list or tuple of weights to match against.

        Returns:
        - A new tensorProduct with a filtered coeff_dict.
        """
        self._compute_weights()

        # Filter components whose weights match the given weight_list
        filtered_coeff_dict = {
            key: value for key, value in self.coeff_dict.items() if self._weights[key] == tuple(weight_list)
        }

        return tensorProduct(self.vector_space, filtered_coeff_dict)

    def tp(self, other):
        """Tensor product with another tensorProduct."""
        if hasattr(other,"algebra") and other.algebra == self.vector_space:
            other = other._convert_to_tp()
        elif hasattr(other,"vectorSpace") and other.vectorSpace == self.vector_space:
            other = other._convert_to_tp()
        if not isinstance(other, tensorProduct):
            raise ValueError("The other object must be a tensorProduct instance.")
        if self.vector_space != other.vector_space:
            raise ValueError("Both tensors must have the same vector_space.")

        # Compute new coefficient dictionary
        new_coeff_dict = {}
        for (key1, value1) in self.coeff_dict.items():
            for (key2, value2) in other.coeff_dict.items():
                new_key = key1[:len(key1)//2] + key2[:len(key2)//2] + key1[len(key1)//2:] + key2[len(key2)//2:]
                new_coeff_dict[new_key] = value1 * value2

        return tensorProduct(self.vector_space, new_coeff_dict)

    def is_zero(self):
        for v in self.coeff_dict.values():
            if sp.simplify(v) != 0:
                return False
        else:
            return True

    def __str__(self):
        return tensor_VS_printer(self)

    def __repr__(self):
        return tensor_VS_printer(self)

    def _latex(self, printer):
        """
        Defines the LaTeX representation for SymPy's latex() function.
        """
        return tensor_latex_helper(self)

    def _sympystr(self, printer):
        return self.__repr__()

    def subs(self, subs_data):
        new_dict = {j:sp.sympify(k).subs(subs_data) for j,k in self.coeff_dict.items()}
        return tensorProduct(self.vector_space,new_dict)

    def simplify(self):
        new_dict = {j:sp.simplify(k) for j,k in self.coeff_dict.items()}
        return tensorProduct(self.vector_space,new_dict)

    def _eval_simplify(self, ratio=1.7, measure=None, rational=True, **kwargs):
        """
        Custom simplification method for sympy.simplify.
        """
        # Apply sympy.simplify to each value in coeff_dict
        new_dict = {key: sp.simplify(value) for key, value in self.coeff_dict.items()}
        # Return a new instance of the class (using tensorProduct or another constructor)
        return tensorProduct(self.vector_space, new_dict)

    def __add__(self,other):
        if hasattr(other,"algebra") and other.algebra == self.vector_space:
            other = other._convert_to_tp()
        elif hasattr(other,"vectorSpace") and other.vectorSpace == self.vector_space:
            other = other._convert_to_tp()
        if not isinstance(other,tensorProduct) and other.vector_space != self.vector_space:
            raise TypeError('`+` requires other tensor to be formed from the same vector space.')
        new_dict = self.coeff_dict
        for key, val in other.coeff_dict.items():
            if key in new_dict:
                new_dict[key] = new_dict[key]+val
            else:
                new_dict[key] = val
        return tensorProduct(self.vector_space,new_dict)

    def __sub__(self,other):
        if hasattr(other,"algebra") and other.algebra == self.vector_space:
            other = other._convert_to_tp()
        elif hasattr(other,"vectorSpace") and other.vectorSpace == self.vector_space:
            other = other._convert_to_tp()
        if not isinstance(other,tensorProduct) and other.vector_space != self.vector_space:
            raise TypeError('`+` requires other tensor to be formed from the same vector space.')
        new_dict = self.coeff_dict
        for key, val in other.coeff_dict.items():
            if key in new_dict:
                new_dict[key] = new_dict[key]-val
            else:
                new_dict[key] = -val
        return tensorProduct(self.vector_space,new_dict)

    def __matmul__(self, other):
        """Overload @ operator for tensor product."""
        return self.tp(other)

    def contract_call(self, other):
        """
        Contract the last index of self with the first index of other or handle AlgebraElement.
        """
        if self.is_zero():
            return self
        if other.is_zero():
            return 0*self
        if isinstance(other, tensorProduct):
            if self.vector_space != other.vector_space:
                raise ValueError("Both tensors must be defined w.r.t. the same vector_space.")

            if self.trailing_valence + other.leading_valence != 1:
                raise ValueError("Contraction requires the first tensor factor of every term in other to have leading valence different from the last entry tensor factor from terms of self.")

            new_dict = {}
            for key1, value1 in self.coeff_dict.items():
                for key2, value2 in other.coeff_dict.items():
                    if key1[(len(key1)//2)-1] == key2[0]:  # Matching indices for contraction
                        new_key = key1[:(len(key1)//2)-1] + key2[1:(len(key2)//2)] + key1[(len(key1)//2):-1] + key2[1+(len(key2)//2):] # Remove the contracted indices
                        new_value = value1 * value2
                        new_dict[new_key] = new_dict.get(new_key, 0) + new_value  # Accumulate values for duplicate keys

            return tensorProduct(self.vector_space, new_dict)

        elif hasattr(other, "algebra") and self.vector_space == other.algebra:
            # Logic for AlgebraElement instances
            if self.trailing_valence != 0:
                raise ValueError("Operating on AlgebraElement requires all terms in self to end in covariant tensor factor.")

            # Convert other (AlgebraElement) into a tensorProduct instance
            other_as_tensor = other._convert_to_tp()

            # Reuse existing contract_call logic
            return self.contract_call(other_as_tensor)

        else:
            raise ValueError("The other object must be a tensorProduct or an AlgebraElement with matching algebra.")

    def _recursion_contract_hom(self,other):
        if self.is_zero():
            return self
        if self.prolongation_type == 0:
            vs = tuple([j.dual() for j in self.vector_space])
            vsDual = self.vector_space.basis
        elif self.prolongation_type == 1:
            vs = self.vector_space.basis
            vsDual = tuple([j.dual() for j in self.vector_space])
        else:
            raise TypeError(f'`_recursion_contract_hom` does not operate on arguments with mixed `prolongation_type` e.g., {self} has type {self.prolongation_type}')
        if self.max_degree==1 or other.max_degree==1:
            return self*other
        otherContract = other*vs[0]
        if hasattr(otherContract,'_convert_to_tp'):
            otherContract=otherContract._convert_to_tp()
        image_part = ((self*vs[0])._recursion_contract_hom(other)+self._recursion_contract_hom(otherContract))
        domain_part = vsDual[0]
        contraction = image_part@domain_part
        for vec,vecD in  zip(vs[1:],vsDual[1:]):
            otherContract = other*vec
            if hasattr(otherContract,'_convert_to_tp'):
                otherContract=otherContract._convert_to_tp()
            contraction += ((self*vec)._recursion_contract_hom(other)+self._recursion_contract_hom(otherContract))@vecD
        return contraction

    def _recursion_contract(self,other):
        if self.is_zero():
            return self
        if other.is_zero():
            return 0*self
        if isinstance(other, tensorProduct):
            if self.prolongation_type != other.prolongation_type:
                raise type(f'`tensorProduct` contraction is only supported between instances with matching `prolongation types`, not types: {self.prolongation_type} and {other.prolongation_type}')
            if self.vector_space != other.vector_space:
                return 0*self
            hc1 = self.homogeneous_components
            hc2 = other.homogeneous_components
            terms = [t1._recursion_contract_hom(t2) for t1,t2 in zip(hc1,hc2)]
            return sum(terms[1:],terms[0])

    def _bracket(self,other):
        if self.is_zero():
            return self
        if other.is_zero():
            return 0*self
        if isinstance(other, tensorProduct):
            if self.vector_space != other.vector_space:
                raise ValueError("In `tensorProduct._bracket` both tensors must be defined w.r.t. the same vector_space.")

            if self.prolongation_type!=other.prolongation_type or self.prolongation_type==-1:
                raise ValueError("`tensorProduct._bracket` requires bracket components to have matching prolongation types.")

            complimentType = 1 if self.prolongation_type==0 else 0

            new_dict = {}
            for key1, value1 in self.coeff_dict.items():
                for key2, value2 in other.coeff_dict.items():
                    if key1[(len(key1)//2)-1] == key2[0]:  # Matching indices for contraction
                        new_value = value1 * value2
                        key1_truncated = list(key1[:(len(key1)//2)-1])
                        key1_start = [key1_truncated[0]]
                        key1_tr_tail = key1_truncated[1:]
                        key2_tail = key2[1:(len(key2)//2)]
                        new_tails = shufflings(key1_tr_tail,key2_tail)
                        valence = [self.prolongation_type]+([complimentType]*(len(key1_tr_tail)+len(key2_tail)))
                        new_keys = [tuple(key1_start+tail+valence) for tail in new_tails]
                        for key in new_keys:
                            new_dict[key] = new_dict.get(key, 0) + new_value  # Accumulate values for duplicate keys
                    if key2[(len(key1)//2)-1] == key1[0]:  # Matching indices for contraction
                        new_value = -value1 * value2
                        key2_truncated = list(key2[:(len(key2)//2)-1])
                        key2_start = [key2_truncated[0]]
                        key2_tr_tail = key2_truncated[1:]
                        key1_tail = key1[1:(len(key1)//2)]
                        new_tails = shufflings(key2_tr_tail,key1_tail)
                        valence = [self.prolongation_type]+([complimentType]*(len(key2_tr_tail)+len(key1_tail)))
                        new_keys = [tuple(key2_start+tail+valence) for tail in new_tails]
                        for key in new_keys:
                            new_dict[key] = new_dict.get(key, 0) + new_value  # Accumulate values for duplicate keys

            return tensorProduct(self.vector_space, new_dict)

        elif hasattr(other, "algebra") and self.vector_space == other.algebra:
            # Logic for AlgebraElement instances
            if self.vector_space != other.algebra:
                raise ValueError("In `tensorProduct._bracket` both tensors must be defined w.r.t. the same vector_space.")

            if self.prolongation_type != other.valence:
                raise ValueError("`tensorProduct._bracket` operating on AlgebraElement requires all terms in self to end in covariant tensor factor.")

            # Convert other (AlgebraElement) into a tensorProduct instance
            other_as_tensor = other._convert_to_tp()
            other_index,other_value = list(other_as_tensor.coeff_dict.items())[0]
            new_dict = {}
            for key, value in self.coeff_dict.items():
                if key[(len(key)//2)-1] == other_index:  # Matching indices for contraction
                    new_value = value * other_value
                    key_truncated = tuple(key[:(len(key)//2)-1]+key[(len(key)//2):-1])
                    new_dict[key_truncated] = new_dict.get(key_truncated, 0) + new_value
            return tensorProduct(self.vector_space, new_dict)

        else:
            raise ValueError("In `tensorProduct._bracket` the second factor must be a tensorProduct or an AlgebraElement with matching algebra.")



    def __mul__(self, other):
        """Overload * to compute the Lie bracket, with special logic for AlgebraElement."""
        if self.max_degree==0:
            return self.coeff_dict[tuple()]*other
        if isinstance(other, (int, float, sp.Expr)):
            new_coeff_dict = {key: value * other for key, value in self.coeff_dict.items()}
            return tensorProduct(self.vector_space, new_coeff_dict)

        if isinstance(other, vectorSpaceElement) or hasattr(other, "algebra"):
            other = other._convert_to_tp()

        if isinstance(other, tensorProduct):
            # Lie bracket for two tensorProducts
            if self.vector_space != other.vector_space:
                return 0*self
            if other.max_degree==0:
                return other.coeff_dict[tuple()]*self

            if other.max_degree==1 and other.min_degree==1:
                if self.max_degree==1 and self.min_degree==1:
                    from DGCV.finiteDimAlgebras import AlgebraElement #local import  # noqa: I001
                    if self.prolongation_type == other.prolongation_type:
                        if isinstance(self.vector_space,vectorSpace):
                            return 0*self
                        pt = self.prolongation_type
                        coeffs1 = [self.coeff_dict.get((j, pt), 0) for j in range(self.vector_space.dimension)]
                        coeffs2 = [other.coeff_dict.get((j, pt), 0) for j in range(self.vector_space.dimension)]
                        LA_elem1 = AlgebraElement(self.vector_space,coeffs1,pt)
                        LA_elem2 = AlgebraElement(other.vector_space,coeffs2,pt)
                        return (LA_elem1*LA_elem2)._convert_to_tp()
                    else:
                        ###!!! mixed prolongation type needs an additional logic branch
                        pt1 = self.prolongation_type
                        pt2 = other.prolongation_type
                        return sum([self.coeff_dict[(j,pt1)]*other.coeff_dict[(j,pt2)] for j in range(self.vector_space.dimension)])
                else:
                    if self.trailing_valence != other.prolongation_type and self.trailing_valence!=-1:
                        cd = {}
                        for t,tv in other.coeff_dict.items():
                            for key,val in {tuple(k[:(len(k)//2)-1]+k[(len(k)//2):-1]):v*tv for k,v in self.coeff_dict.items() if k[(len(k)//2)-1]==t[0]}.items():
                                cd[key] = cd.get(key,0)+val
                        return tensorProduct(self.vector_space,cd)
                    else:
                        raise TypeError(f'* cannot operate on `tensorProduct` pairs if: \n second arg is degree 1 \n first argument has degree>1 \n first\'s `trailing_valence` doesn\'t match the second\'s `prolongation_type`.\n `trailing_valence` of {self}: {self.trailing_valence} \n `prolongation_type` of {other}: {other.prolongation_type}')
            if self.max_degree==1 and self.min_degree==1:
                return -1 * other * self
            return self._bracket(other)
        else:
            raise ValueError(f"Unsupported operation for * between the given object types: {type(self)} and {type(other)}")

    def __rmul__(self, other):
        """Right-hand multiplication for AlgebraElement * tensorProduct."""
        if isinstance(other, (int, float, sp.Expr)):
            return self.__mul__(other)
        if hasattr(other, "algebra") and other.algebra == self.vector_space:
            # Reverse order: AlgebraElement * tensorProduct
            return -self.contract_call(other)
        else:
            return NotImplemented


################ creator functions
def createVectorSpace(
    obj,
    label,
    basis_labels=None,
    grading=None,
    verbose=False
):
    """
    Registers a vector space object and its basis elements in the caller's global namespace,
    and adds them to the variable_registry for tracking in the Variable Management Framework.

    Parameters
    ----------
    obj : int, vectorSpace, list of vectorSpaceElements
        vector space dimension
    label : str
        The label used to reference the VS object in the global namespace.
    basis_labels : list, optional
        A list of custom labels for the basis elements of the VS.
        If not provided, default labels will be generated.
    grading : list of lists or list, optional
        A list specifying the grading(s) of the VS.
    verbose : bool, optional
        If True, provides detailed feedback during the creation process.
    """

    if label in listVar(algebras_only=True):
        warnings.warn('`createFiniteAlg` was called with a `label` already assigned to another algebra, so `createFiniteAlg` will overwrite the other algebra.')
        clearVar(label)

    # Validate or create the vectorSpace object
    if isinstance(obj, vectorSpace):
        if verbose:
            print(f"Using existing vectorSpace instance: {label}")
        dimension = obj.dimension
    elif isinstance(obj, list) and all(isinstance(el, vectorSpaceElement) for el in obj):
        if verbose:
            print("Creating VS from list of vectorSpaceElement instances.")
        dimension = len(obj)
    elif isinstance(obj,int) and 0<=obj:
        dimension = obj

    # Create or validate basis labels
    if basis_labels is None:
        basis_labels = [validate_label(f"{label}_{i+1}") for i in range(dimension)]
    validate_label_list(basis_labels)

    # Process grading
    if grading is None:
        grading = [(0,) * dimension]  # Default grading: all zeros
    elif isinstance(grading, (list, tuple)) and all(
        isinstance(w, (int, sp.Expr)) for w in grading
    ):
        # Single grading vector
        if len(grading) != dimension:
            raise ValueError(
                f"Grading vector length ({len(grading)}) must match the VS dimension ({dimension})."
            )
        grading = [tuple(grading)]  # Wrap single vector in a list
    elif isinstance(grading, list) and all(
        isinstance(vec, (list, tuple)) for vec in grading
    ):
        # List of grading vectors
        for vec in grading:
            if len(vec) != dimension:
                raise ValueError(
                    f"Grading vector length ({len(vec)}) must match the VS dimension ({dimension})."
                )
        grading = [tuple(vec) for vec in grading]  # Convert each vector to Tuple
    else:
        raise ValueError("Grading must be a single vector or a list of vectors.")

    # Create the vectorSpace object
    passkey = retrieve_passkey()

    vs_obj = vectorSpace(
        dimension,
        grading=grading,
        _label=label,
        _basis_labels=basis_labels,
        _calledFromCreator=passkey,
    )

    # initialize vector space and its basis
    assert (
        vs_obj.basis is not None
    ), "VS object basis elements must be initialized."

    # Register in _cached_caller_globals
    _cached_caller_globals.update({label: vs_obj})
    _cached_caller_globals.update(zip(basis_labels, vs_obj.basis))

    # Register in the variable registry
    variable_registry = get_variable_registry()
    variable_registry["finite_algebra_systems"][label] = {
        "family_type": "algebra",
        "family_names": tuple(basis_labels),
        "family_values": tuple(vs_obj.basis),
        "dimension": dimension,
        "grading": grading,
        "basis_labels": basis_labels,
        "structure_data": {(0,0,0):0},
    }

    if verbose:
        print(f"Vector Space '{label}' registered successfully.")
        print(
            f"Dimension: {dimension}, Grading: {grading}, Basis Labels: {basis_labels}"
        )

    return vs_obj





